from typing import MutableMapping, Optional, Union

from tair.typing import ResponseCallbackT
from tair.commands import TairCommands

from redis import ConnectionPool
from redis.asyncio.client import Pipeline as RedisPipeline


class Pipeline(RedisPipeline, TairCommands):
    def __init__(
        self,
        connection_pool: ConnectionPool,
        response_callbacks: MutableMapping[Union[str, bytes], ResponseCallbackT],
        transaction: bool,
        shard_hint: Optional[str],
    ):
        RedisPipeline.__init__(
            self,
            connection_pool=connection_pool,
            response_callbacks=response_callbacks,
            transaction=transaction,
            shard_hint=shard_hint,
        )



from typing import Union

from .tairhash import (
    TairHashCommands,
    parse_exhincrbyfloat,
    parse_exhgetwithver,
    parse_exhmgetwithver,
    parse_exhgetall,
    parse_exhscan,
)

from .tairstring import (
    TairStringCommands,
    parse_exset,
    parse_exget,
    parse_excas,
    parse_exincrbyfloat,
)

from .tairzset import TairZsetCommands, parse_tair_zset_items


from redis import Redis
from redis.asyncio import Redis as AsyncRedis


class TairCommands(
    TairHashCommands,
    TairStringCommands,
    TairZsetCommands,
):
    pass


def str_if_bytes(value: Union[str, bytes]) -> str:
    return (
        value.decode("utf-8", errors="replace") if isinstance(value, bytes) else value
    )


def bool_ok(resp) -> bool:
    return str_if_bytes(resp) == "OK"


TAIR_RESPONSE_CALLBACKS = {
    # TairString
    "EXSET": parse_exset,
    "EXGET": parse_exget,
    "EXINCRBYFLOAT": parse_exincrbyfloat,
    "EXCAS": parse_excas,
    # TairHash
    "EXHMSET": bool_ok,
    "EXHINCRBYFLOAT": parse_exhincrbyfloat,
    "EXHGETWITHVER": parse_exhgetwithver,
    "EXHMGETWITHVER": parse_exhmgetwithver,
    "EXHGETALL": parse_exhgetall,
    "EXHSCAN": parse_exhscan,
    # TairZset
    "EXZINCRBY": str_if_bytes,
    "EXZSCORE": str_if_bytes,
    "EXZRANGE": parse_tair_zset_items,
    "EXZREVRANGE": parse_tair_zset_items,
    "EXZRANGEBYSCORE": parse_tair_zset_items,
    "EXZREVRANGEBYSCORE": parse_tair_zset_items,
}


def set_tair_response_callback(redis: Union[Redis, AsyncRedis]):
    for cmd, cb in TAIR_RESPONSE_CALLBACKS.items():
        redis.set_response_callback(cmd, cb)



from .client import Tair
from .cluster import TairCluster

from .tairstring import ExgetResult, ExcasResult
from .tairhash import ExhscanResult, FieldValueItem, ValueVersionItem
from .tairzset import TairZsetItem

from .exceptions import (
    AuthenticationError,
    AuthenticationWrongNumberOfArgsError,
    BusyLoadingError,
    ChildDeadlockedError,
    ConnectionError,
    DataError,
    InvalidResponse,
    PubSubError,
    ReadOnlyError,
    TairError,
    ResponseError,
    TimeoutError,
    WatchError,
)


__all__ = [
    "Tair",
    "TairCluster",
    "ExgetResult",
    "ExcasResult",
    "ExhscanResult",
    "FieldValueItem",
    "ValueVersionItem",
    "TairZsetItem",
    "AuthenticationError",
    "AuthenticationWrongNumberOfArgsError",
    "BusyLoadingError",
    "ChildDeadlockedError",
    "ConnectionError",
    "DataError",
    "InvalidResponse",
    "PubSubError",
    "ReadOnlyError",
    "ResponseError",
    "TairError",
    "TimeoutError",
    "WatchError",
]



from typing import List, Optional

from tair.commands import TairCommands, set_tair_response_callback

from redis.asyncio.cluster import RedisCluster, ClusterNode


class TairCluster(RedisCluster, TairCommands):
    @classmethod
    def from_url(cls, url, **kwargs):
        return cls(url=url, **kwargs)

    def __init__(
        self,
        host: Optional[str] = None,
        port: int = 6379,
        startup_nodes: Optional[List[ClusterNode]] = None,
        require_full_coverage: bool = False,
        read_from_replicas: bool = False,
        cluster_error_retry_attempts: int = 3,
        reinitialize_steps: int = 10,
        url: Optional[str] = None,
        **kwargs,
    ):
        RedisCluster.__init__(
            self,
            host=host,
            port=port,
            startup_nodes=startup_nodes,
            require_full_coverage=require_full_coverage,
            read_from_replicas=read_from_replicas,
            cluster_error_retry_attempts=cluster_error_retry_attempts,
            reinitialize_steps=reinitialize_steps,
            url=url,
            **kwargs,
        )
        set_tair_response_callback(self)

    async def initialize(self):
        await RedisCluster.initialize(self)
